#!/usr/bin/env python3
#
#  whey.py
"""
Parser for whey's own configuration.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Dict, List, Set, Type, cast

# 3rd party
import dist_meta.entry_points
from dom_toml.parser import TOML_TYPES, AbstractConfigParser, BadConfigError
from natsort import natsorted
from shippinglabel.classifiers import validate_classifiers

# this package
from whey import additional_files
from whey.builder import AbstractBuilder, SDistBuilder, WheelBuilder

__all__ = (
		"WheyParser",
		"backfill_classifiers",
		"get_default_builders",
		"get_entry_points",
		"license_lookup",
		)

#: Mapping of license short codes to license names used in trove classifiers.
license_lookup = {
		"Apache-2.0": "Apache Software License",
		"BSD": "BSD License",
		"BSD-2-Clause": "BSD License",
		"BSD-3-Clause": "BSD License",
		"AGPL-3.0-only": "GNU Affero General Public License v3",
		"AGPL-3.0": "GNU Affero General Public License v3",
		"AGPL-3.0-or-later": "GNU Affero General Public License v3 or later (AGPLv3+)",
		"AGPL-3.0+": "GNU Affero General Public License v3 or later (AGPLv3+)",
		"FDL": "GNU Free Documentation License (FDL)",
		"GFDL-1.1-only": "GNU Free Documentation License (FDL)",
		"GFDL-1.1-or-later": "GNU Free Documentation License (FDL)",
		"GFDL-1.2-only": "GNU Free Documentation License (FDL)",
		"GFDL-1.2-or-later": "GNU Free Documentation License (FDL)",
		"GFDL-1.3-only": "GNU Free Documentation License (FDL)",
		"GFDL-1.3-or-later": "GNU Free Documentation License (FDL)",
		"GFDL-1.2": "GNU Free Documentation License (FDL)",
		"GFDL-1.1": "GNU Free Documentation License (FDL)",
		"GFDL-1.3": "GNU Free Documentation License (FDL)",
		"GPL": "GNU General Public License (GPL)",
		"GPL-1.0-only": "GNU General Public License (GPL)",
		"GPL-1.0-or-later": "GNU General Public License (GPL)",
		"GPLv2": "GNU General Public License v2 (GPLv2)",
		"GPL-2.0-only": "GNU General Public License v2 (GPLv2)",
		"GPLv2+": "GNU General Public License v2 or later (GPLv2+)",
		"GPL-2.0-or-later": "GNU General Public License v2 or later (GPLv2+)",
		"GPLv3": "GNU General Public License v3 (GPLv3)",
		"GPL-3.0-only": "GNU General Public License v3 (GPLv3)",
		"GPLv3+": "GNU General Public License v3 or later (GPLv3+)",
		"GPL-3.0-or-later": "GNU General Public License v3 or later (GPLv3+)",
		"LGPLv2": "GNU Lesser General Public License v2 (LGPLv2)",
		"LGPLv2+": "GNU Lesser General Public License v2 or later (LGPLv2+)",
		"LGPLv3": "GNU Lesser General Public License v3 (LGPLv3)",
		"LGPL-3.0-only": "GNU Lesser General Public License v3 (LGPLv3)",
		"LGPLv3+": "GNU Lesser General Public License v3 or later (LGPLv3+)",
		"LGPL-3.0-or-later": "GNU Lesser General Public License v3 or later (LGPLv3+)",
		"LGPL": "GNU Library or Lesser General Public License (LGPL)",
		"MIT": "MIT License",
		"PSF-2.0": "Python Software Foundation License",
		}


def get_default_builders() -> Dict[str, Type[AbstractBuilder]]:
	"""
	Returns a mapping of builder categories to builder classes to use as the default builders.
	"""

	return {"sdist": SDistBuilder, "binary": WheelBuilder, "wheel": WheelBuilder}


class WheyParser(AbstractConfigParser):
	"""
	Parser for the ``[tool.whey]`` table from ``pyproject.toml``.

	.. autosummary-widths:: 1/2
	"""

	defaults = {
			"source-dir": '.',
			"license-key": None,
			"platforms": None,
			"python-versions": None,
			"python-implementations": None,
			}
	factories = {
			"additional-files": list,
			"base-classifiers": list,
			"builders": get_default_builders,
			}

	def parse_package(self, config: Dict[str, TOML_TYPES]) -> str:
		"""
		Parse the ``package`` key, giving the name of the importable package.

		This defaults to :pep621:`project.name <name>` if unspecified.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		package = config["package"]

		self.assert_type(package, str, ["tool", "whey", "package"])

		return package

	def parse_source_dir(self, config: Dict[str, TOML_TYPES]) -> str:
		"""
		Parse the ``source-dir`` key, giving the name of the directory containing the project's source.

		This defaults to ``'.'`` if unspecified.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		source_dir = config["source-dir"]

		self.assert_type(source_dir, str, ["tool", "whey", "source-dir"])

		return source_dir

	def parse_license_key(self, config: Dict[str, TOML_TYPES]) -> str:
		"""
		Parse the ``license-key`` key, giving the identifier of the project's license. Optional.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		license_key = config["license-key"]

		self.assert_type(license_key, str, ["tool", "whey", "license-key"])

		return license_key

	def parse_additional_files(self, config: Dict[str, TOML_TYPES]) -> List[additional_files.AdditionalFilesEntry]:
		"""
		Parse the ``additional-files`` key, giving `MANIFEST.in`_-style
		entries for additional files to include in distributions.

		.. _MANIFEST.in: https://packaging.python.org/guides/using-manifest-in/

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""  # noqa: D400

		entries = config["additional-files"]

		for idx, file in enumerate(entries):
			self.assert_indexed_type(file, str, ["tool", "whey", "additional-files"], idx=idx)

		parsed_additional_files = []

		for entry in entries:
			parsed_entry = additional_files.from_entry(entry)
			if parsed_entry is not None:
				parsed_additional_files.append(parsed_entry)

		return parsed_additional_files

	def parse_platforms(self, config: Dict[str, TOML_TYPES]) -> List[str]:
		"""
		Parse the ``platforms`` key, giving a list of supported platforms. Optional.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		platforms = config["platforms"]

		for idx, plat in enumerate(platforms):
			self.assert_indexed_type(plat, str, ["tool", "whey", "platforms"], idx=idx)

		return platforms

	@staticmethod
	def parse_python_versions(config: Dict[str, TOML_TYPES]) -> List[str]:
		"""
		Parse the ``python-versions`` key, giving a list of supported Python versions. Optional.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		python_versions = config["python-versions"]

		for idx, version in enumerate(python_versions):
			if not isinstance(version, (str, int, float)):
				raise TypeError(
						f"Invalid type for 'tool.whey.python-versions[{idx}]': "
						f"expected {str!r}, {int!r} or {float!r}, got {type(version)!r}"
						)
			if str(version)[0] in "12":
				raise BadConfigError(
						f"Invalid value for 'tool.whey.python-versions[{idx}]': whey only supports Python 3-only projects."
						)

		return list(map(str, python_versions))

	def parse_python_implementations(self, config: Dict[str, TOML_TYPES]) -> List[str]:
		"""
		Parse the ``python-implementations`` key, giving a list of supported Python implementations. Optional.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		python_implementations = config["python-implementations"]

		for idx, impl in enumerate(python_implementations):
			self.assert_indexed_type(impl, str, ["tool", "whey", "python-implementations"], idx=idx)

		return python_implementations

	def parse_base_classifiers(self, config: Dict[str, TOML_TYPES]) -> Set[str]:
		"""
		Parse the ``base-classifiers`` key, giving a list `trove classifiers <https://pypi.org/classifiers/>`__.

		This list will be extended with the appropriate classifiers for supported platforms,
		Python versions and implementations, and the project's license.
		Ignored if :pep621:`classifiers` is not listed in :pep621:`dynamic`

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.

		:rtype:

		.. latex:clearpage::
		"""

		parsed_classifiers = set()

		for idx, classifier in enumerate(config["base-classifiers"]):
			self.assert_indexed_type(classifier, str, ["tool", "whey", "python-implementations"], idx=idx)
			parsed_classifiers.add(classifier)

		return parsed_classifiers

	def parse_builders(self, config: Dict[str, TOML_TYPES]) -> Dict[str, Type[AbstractBuilder]]:
		"""
		Parse the ``builders`` table, which lists gives the entry points to use for the sdist and wheel builders.

		This allows the user to select a custom builder with additional functionality.

		:param config: The unparsed TOML config for the ``[tool.whey]`` table.
		"""

		parsed_builders = get_default_builders()
		builders = config["builders"]

		entry_points: Dict[str, dist_meta.entry_points.EntryPoint] = get_entry_points()

		self.assert_type(builders, dict, ["tool", "whey", "builders"])

		for builder_type in ["binary", "sdist", "wheel"]:
			if builder_type in builders:
				entry_point_name = builders[builder_type]
				if entry_point_name not in entry_points:
					raise BadConfigError(
							f"Unknown {builder_type} builder {entry_point_name!r}. \n"
							f"Is it registered as an entry point under 'whey.builder'?"
							)

				parsed_builders[builder_type] = cast(Type[AbstractBuilder], entry_points[entry_point_name].load())

		return parsed_builders

	@property
	def keys(self) -> List[str]:
		"""
		The keys to parse from the TOML file.
		"""

		return [
				"package",
				"source-dir",
				"additional-files",
				"license-key",
				"base-classifiers",
				"platforms",
				"python-versions",
				"python-implementations",
				"builders",
				]


def backfill_classifiers(config: Dict[str, TOML_TYPES]) -> List[str]:
	"""
	Backfill `trove classifiers <https://pypi.org/classifiers/>`_ for supported platforms,
	Python versions and implementations, and the project's license, as appropriate.

	:param config: The parsed config from ``pyproject.toml``.
	"""  # noqa: D400

	# TODO: Typing :: Typed

	parsed_classifiers = set(config["base-classifiers"])

	platforms = config["platforms"]
	license_key = config["license-key"]
	python_versions = config["python-versions"]
	python_implementations = config["python-implementations"]

	if license_key in license_lookup:
		parsed_classifiers.add(f"License :: OSI Approved :: {license_lookup[license_key]}")

	if platforms:

		if set(platforms) == {"Windows", "macOS", "Linux"}:
			parsed_classifiers.add("Operating System :: OS Independent")
		else:
			if "Windows" in platforms:
				parsed_classifiers.add("Operating System :: Microsoft :: Windows")
			if "Linux" in platforms:
				parsed_classifiers.add("Operating System :: POSIX :: Linux")
			if "macOS" in platforms:
				parsed_classifiers.add("Operating System :: MacOS")

	if python_versions:
		for version in python_versions:
			parsed_classifiers.add(f"Programming Language :: Python :: {version}")

		parsed_classifiers.add("Programming Language :: Python :: 3 :: Only")

	if python_implementations:
		for implementation in python_implementations:
			parsed_classifiers.add(f"Programming Language :: Python :: Implementation :: {implementation}")

	parsed_classifiers.add("Programming Language :: Python")

	validate_classifiers(parsed_classifiers)

	return natsorted(parsed_classifiers)


def get_entry_points(group: str = "whey.builder") -> Dict[str, dist_meta.entry_points.EntryPoint]:
	r"""
	Returns an iterable over `EntryPoint`_ objects in the ``group`` group.

	:param group:

	:rtype: :class:`Iterable <typing.Iterable>`\[`EntryPoint`_\]

	.. _EntryPoint: https://docs.python.org/3/library/importlib.metadata.html#entry-points
	"""

	eps = dist_meta.entry_points.get_entry_points(group)

	entry_points: Dict[str, dist_meta.entry_points.EntryPoint] = {}

	for entry_point in eps:  # pylint: disable=use-dict-comprehension
		if entry_point.group == group:
			entry_points[entry_point.name] = entry_point

	return entry_points



#!/usr/bin/env python3
#
#  foreman.py
"""
The foreman is responsible for loading the configuration calling the builders.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Optional, Type

# 3rd party
from consolekit.terminal_colours import ColourTrilean
from domdf_python_tools.paths import PathPlus, traverse_to_file
from domdf_python_tools.typing import PathLike

# this package
from whey.builder import AbstractBuilder
from whey.config import load_toml

__all__ = ("Foreman", )


class Foreman:
	"""
	Responsible for loading the configuration calling the builders.
	"""

	def __init__(self, project_dir: PathLike):

		# Walk up the tree until a ``pyproject.toml`` file is found.
		#: The pyproject.toml directory
		self.project_dir: PathPlus = traverse_to_file(PathPlus(project_dir), "pyproject.toml")

		#: Configuration parsed from ``pyproject.toml``.
		self.config = load_toml(self.project_dir / "pyproject.toml")

	def get_builder(self, distribution_type: str) -> Type[AbstractBuilder]:
		"""
		Returns the builder for the given distribution type.

		:param distribution_type: The distribution type, such as ``'source'`` or ``'wheel'``.
		"""

		return self.config["builders"][distribution_type]

	def build_sdist(
			self,
			build_dir: Optional[PathLike] = None,
			out_dir: Optional[PathLike] = None,
			*args,
			verbose: bool = False,
			colour: ColourTrilean = None,
			**kwargs,
			) -> str:
		"""
		Build a sdist distribution using the ``sdist`` builder configured in ``pyproject.toml``.

		:returns: The filename of the created archive.
		"""

		builder = self.get_builder("sdist")(
				self.project_dir,
				self.config,
				build_dir,
				out_dir,
				*args,
				verbose=verbose,
				colour=colour,
				**kwargs,
				)
		return builder.build()

	def build_binary(
			self,
			build_dir: Optional[PathLike] = None,
			out_dir: Optional[PathLike] = None,
			*args,
			verbose: bool = False,
			colour: ColourTrilean = None,
			**kwargs,
			) -> str:
		"""
		Build a binary distribution using the ``binary`` builder configured in ``pyproject.toml``.

		:returns: The filename of the created archive.
		"""

		builder = self.get_builder("binary")(
				self.project_dir,
				self.config,
				build_dir,
				out_dir,
				*args,
				verbose=verbose,
				colour=colour,
				**kwargs,
				)
		return builder.build()

	def build_wheel(
			self,
			build_dir: Optional[PathLike] = None,
			out_dir: Optional[PathLike] = None,
			*args,
			verbose: bool = False,
			colour: ColourTrilean = None,
			**kwargs,
			) -> str:
		"""
		Build a wheel distribution using the ``wheel`` builder configured in ``pyproject.toml``.

		:returns: The filename of the created archive.
		"""

		builder = self.get_builder("wheel")(
				self.project_dir,
				self.config,
				build_dir,
				out_dir,
				*args,
				verbose=verbose,
				colour=colour,
				**kwargs,
				)
		return builder.build()



#!/usr/bin/env python3
#
#  utils.py
"""
General utilities.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Dict, Iterable, Optional, Type, cast

# 3rd party
import click
from domdf_python_tools.words import Plural
from pyproject_parser.cli import ConfigTracebackHandler

# this package
from whey.builder import AbstractBuilder
from whey.config.whey import get_entry_points
from whey.foreman import Foreman

__all__ = ("WheyTracebackHandler", "parse_custom_builders", "print_builder_names")

_builder = Plural("builder", "builders")


class WheyTracebackHandler(ConfigTracebackHandler):
	"""
	Custom :class:`consolekit.tracebacks.TracebackHandler`.
	"""


class WheyBackendTBHandler(WheyTracebackHandler):

	def __init__(self, exception: BaseException = SystemExit(1)):
		super().__init__(exception)


def parse_custom_builders(builders: Optional[Iterable[str]]) -> Dict[str, Type[AbstractBuilder]]:
	"""
	Parse the custom builders passed using the ``--builder NAME`` option on the command line.

	:param builders:
	"""

	custom_builders: Dict[str, Type[AbstractBuilder]] = {}

	if builders is None:
		return custom_builders

	entry_points = get_entry_points()
	for builder_name in builders:
		if builder_name not in entry_points:
			raise click.BadArgumentUsage(
					f"Unknown builder {builder_name!r}. \n"
					f"Is it registered as an entry point under 'whey.builder'?"
					)
		else:
			custom_builders[builder_name] = cast(Type[AbstractBuilder], entry_points[builder_name].load())

	return custom_builders


def print_builder_names(
		foreman: Foreman,
		custom_builders: Dict[str, Type[AbstractBuilder]],
		**opts: bool,
		) -> None:
	"""
	Prints the name(s) of the builders which will be used.

	:param foreman:
	:param custom_builders:
	:param opts: Keyword arguments for builder categories.
		Corresponds to the ``--sdist``, ``--wheel`` and ``--binary`` options.
	"""

	builders = []

	for builder_name, enabled in opts.items():
		if enabled:
			builder_obj = foreman.config["builders"][builder_name]  # pylint: disable=loop-invariant-statement
			builders.append(f"    {builder_name}: {builder_obj.__module__}.{builder_obj.__qualname__}")

	for builder_name, builder_obj in custom_builders.items():  # pylint: disable=use-list-copy
		builders.append(f"    {builder_name}: {builder_obj.__module__}.{builder_obj.__qualname__}")

	click.echo(f"Using the following {_builder(len(builders))}:")

	for line in builders:
		click.echo(line)

	click.echo()



#!/usr/bin/env python3
#
#  __init__.py
"""
``pyproject.toml`` configuration parsing.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from typing import Any, Dict

# 3rd party
import dom_toml
import dom_toml.decoder
from dom_toml.parser import BadConfigError
from domdf_python_tools.iterative import natmin
from domdf_python_tools.paths import PathPlus, in_directory
from domdf_python_tools.typing import PathLike
from packaging.requirements import InvalidRequirement
from packaging.specifiers import Specifier
from shippinglabel.requirements import combine_requirements, read_requirements

# this package
from whey.config.pep621 import PEP621Parser
from whey.config.whey import WheyParser, backfill_classifiers

__all__ = (
		"PEP621Parser",
		"WheyParser",
		"backfill_classifiers",
		"load_toml",
		)

_name_to_package_re = re.compile("-(?!stubs)")


def _get_default_package(name: str) -> str:
	return _name_to_package_re.sub('_', name.split('.', 1)[0])


def load_toml(filename: PathLike) -> Dict[str, Any]:  # TODO: TypedDict
	"""
	Load the ``whey`` configuration mapping from the given TOML file.

	:param filename:
	"""

	filename = PathPlus(filename)

	project_dir = filename.parent
	config = dom_toml.load(filename, decoder=dom_toml.decoder.TomlPureDecoder)

	parsed_config = {}
	tool_table = config.get("tool", {})

	with in_directory(project_dir):

		parsed_config.update(WheyParser().parse(tool_table.get("whey", {}), set_defaults=True))

		if "project" in config:
			parsed_config.update(PEP621Parser().parse(config["project"], set_defaults=True))
		else:
			raise KeyError(f"'project' table not found in '{filename!s}'")

		if parsed_config.get("readme", None) is not None:
			parsed_config["readme"] = parsed_config["readme"].resolve()

		if parsed_config.get("license", None) is not None:
			parsed_config["license"] = parsed_config["license"].resolve()

	# set defaults
	parsed_config.setdefault("package", _get_default_package(config["project"]["name"]))

	dynamic_fields = parsed_config.get("dynamic", [])

	if "classifiers" in dynamic_fields:
		dynamic_fields.remove("classifiers")
		parsed_config["classifiers"] = backfill_classifiers(parsed_config)

	if "requires-python" in dynamic_fields:
		if parsed_config["python-versions"]:
			dynamic_fields.remove("requires-python")
			parsed_config["requires-python"] = Specifier(f">={natmin(parsed_config['python-versions'])}")
		else:
			raise BadConfigError(
					"'requires-python' was listed in 'project.dynamic', "
					"but whey cannot determine the minimum supported Python version. \n"
					"Set 'tool.whey.python-versions' to a list of supported Python versions to fix this."
					)

	if "dependencies" in dynamic_fields:
		dynamic_fields.remove("dependencies")

		if (project_dir / "requirements.txt").is_file():
			dependencies, comments, invalid = read_requirements(project_dir / "requirements.txt", include_invalid=True)

			for bad_string in invalid:
				raise InvalidRequirement(bad_string)

			parsed_config["dependencies"] = sorted(combine_requirements(dependencies))
		else:
			raise BadConfigError(
					"'project.dependencies' was listed as a dynamic field "
					"but no 'requirements.txt' file was found."
					)

	parsed_config["dynamic"] = dynamic_fields

	if "base-classifiers" in parsed_config:
		del parsed_config["base-classifiers"]

	return parsed_config



#!/usr/bin/env python
#
#  builder.py
"""
The actual wheel builder.
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import datetime
import os
import pathlib
import re
import shutil
import sys
import tarfile
import zipfile
from abc import ABC, abstractmethod
from email.headerregistry import Address
from functools import partial
from posixpath import join as posixpath_join
from typing import TYPE_CHECKING, Any, Dict, Iterator, Mapping, Optional
from warnings import warn as warnings_warn

# 3rd party
import click
import dom_toml
import handy_archives
from consolekit.terminal_colours import ColourTrilean, Fore, resolve_color_default
from dist_meta import entry_points, metadata, wheel
from dist_meta.metadata_mapping import MetadataMapping
from domdf_python_tools.paths import PathPlus, sort_paths, traverse_to_file
from domdf_python_tools.typing import PathLike
from domdf_python_tools.words import word_join
from pyproject_parser import PyProjectTomlEncoder
from shippinglabel.checksum import get_record_entry
from shippinglabel.requirements import ComparableRequirement, combine_requirements

# this package
from whey import additional_files

__all__ = ("AbstractBuilder", "SDistBuilder", "WheelBuilder")

archive_name_sub_re = re.compile(
		r"[^\w\d.]+",
		re.UNICODE,
		)


class AbstractBuilder(ABC):
	"""
	Abstract base class for builders of Python distributions using metadata read from ``pyproject.toml``.

	:param project_dir: The project to build the distribution for.
	:param build_dir: The (temporary) build directory.
	:default build_dir: :file:`{<project_dir>}/build/`
	:param out_dir: The output directory.
	:default out_dir: :file:`{<project_dir>}/dist/`
	:param verbose: Whether to enable verbose output.
	:param colour: Whether to use coloured output.

	.. autosummary-widths:: 1/2

	.. autoclasssumm:: AbstractBuilder
		:autosummary-sections: Attributes

	.. latex:clearpage::

	.. autosummary-widths:: 7/16

	.. autoclasssumm:: AbstractBuilder
		:autosummary-sections: Methods
	"""

	def __init__(
			self,
			project_dir: PathPlus,
			config: Mapping[str, Any],
			build_dir: Optional[PathLike] = None,
			out_dir: Optional[PathLike] = None,
			*args,
			verbose: bool = False,
			colour: ColourTrilean = None,
			**kwargs,
			):

		# Walk up the tree until a "pyproject.toml" file is found.
		#: The pyproject.toml directory
		self.project_dir: PathPlus = traverse_to_file(PathPlus(project_dir), "pyproject.toml")

		#: Configuration parsed from ``pyproject.toml``.
		self.config: Dict[str, Any] = dict(config)

		#: The archive name, without the tag
		self.archive_name = archive_name_sub_re.sub(
				'_',
				self.config["name"],
				) + f"-{self.config['version']}"

		#: The (temporary) build directory.
		self.build_dir = PathPlus(build_dir or self.default_build_dir)
		self.clear_build_dir()

		#: The output directory.
		self.out_dir = PathPlus(out_dir or self.default_out_dir)
		self.out_dir.maybe_make(parents=True)

		#: Whether to enable verbose output.
		self.verbose = verbose

		#: Whether to use coloured output.
		self.colour = resolve_color_default(colour)

		self._echo = partial(click.echo, color=self.colour)

	@property
	def default_build_dir(self) -> PathPlus:  # pragma: no cover
		"""
		Provides a default for the ``build_dir`` argument.
		"""

		return self.project_dir / "build"

	@property
	def default_out_dir(self) -> PathPlus:  # pragma: no cover
		"""
		Provides a default for the ``out_dir`` argument.
		"""

		return self.project_dir / "dist"

	@property
	def code_directory(self) -> str:
		"""
		The directory containing the code in the build directory.
		"""

		return self.config["source-dir"]

	def clear_build_dir(self) -> None:
		"""
		Clear the build directory of any residue from previous builds.
		"""

		if self.build_dir.is_dir():
			shutil.rmtree(self.build_dir)

		self.build_dir.maybe_make(parents=True)

	def iter_source_files(self) -> Iterator[PathPlus]:
		"""
		Iterate over the files in the source directory.
		"""

		pkgdir = self.project_dir / self.config["source-dir"] / self.config["package"]

		if not pkgdir.is_dir():
			message = f"Package directory {self.config['package']!r} not found"

			if self.config["source-dir"]:
				raise FileNotFoundError(f"{message} in {self.config['source-dir']!r}.")
			else:
				raise FileNotFoundError(f"{message}.")

		found_file = False

		for py_pattern in {"**/*.py", "**/*.pyi", "**/*.pyx", "**/py.typed"}:
			for py_file in pkgdir.rglob(py_pattern):
				if "__pycache__" not in py_file.parts:
					found_file = True
					yield py_file

		if not found_file:
			raise FileNotFoundError(f"No Python source files found in {pkgdir}")

	def copy_source(self) -> None:
		"""
		Copy source files into the build directory.
		"""

		for py_file in self.iter_source_files():
			target = self.build_dir / py_file.relative_to(self.project_dir / self.code_directory)
			target.parent.maybe_make(parents=True)
			target.write_clean(py_file.read_text())
			shutil.copystat(py_file, target)  # pylint: disable=dotted-import-in-loop
			self.report_copied(py_file, target)

	def _echo_if_v(self, *args, **kwargs) -> None:
		if self.verbose:
			self._echo(*args, **kwargs)

	def report_copied(self, source: pathlib.Path, target: pathlib.Path) -> None:
		"""
		Report that a file has been copied into the build directory.

		The format is::

			Copying {source} -> {target.relative_to(self.build_dir)}

		.. latex:vspace:: -5px

		:param source: The source file
		:param target: The file in the build directory.
		"""

		self._echo_if_v(
				f"Copying {source.resolve().as_posix()} -> {target.relative_to(self.build_dir).as_posix()}"
				)

	def report_removed(self, removed_file: pathlib.Path) -> None:
		"""
		Reports the removal of a file from the build directory.

		The format is::

			Removing {removed_file.relative_to(self.build_dir)}

		.. latex:vspace:: -5px

		:param removed_file:
		"""

		self._echo_if_v(f"Removing {removed_file.relative_to(self.build_dir).as_posix()}")

	def report_written(self, written_file: pathlib.Path) -> None:
		"""
		Report that a file has been written to the build directory.

		The format is::

			Writing {written_file.relative_to(self.build_dir)}

		.. latex:vspace:: -5px

		:param written_file:
		"""

		self._echo_if_v(f"Writing {written_file.relative_to(self.build_dir).as_posix()}")

	def copy_additional_files(self) -> None:
		"""
		Copy additional files to the build directory, as specified in the ``additional-files`` key.
		"""

		self.parse_additional_files(*self.config["additional-files"])

	def parse_additional_files(self, *entries: additional_files.AdditionalFilesEntry) -> None:  # pylint: disable=useless-return
		r"""
		Copy additional files to the build directory, by parsing `MANIFEST.in`_-style entries.

		.. _MANIFEST.in: https://packaging.python.org/guides/using-manifest-in/

		:param \*entries:
		"""

		# this package
		from whey import additional_files

		for entry in entries:
			if isinstance(entry, (additional_files.Include, additional_files.RecursiveInclude)):
				for include_file in entry.iter_files(self.project_dir):
					target = self.build_dir / include_file.relative_to(self.project_dir / self.code_directory)
					target.parent.maybe_make(parents=True)
					shutil.copy2(src=include_file, dst=target)
					self.report_copied(include_file, target)
			elif isinstance(entry, (additional_files.Exclude, additional_files.RecursiveExclude)):
				for exclude_file in entry.iter_files(self.build_dir):
					exclude_file.unlink()
					self.report_removed(exclude_file)
			else:  # pragma: no cover
				warnings_warn(f"Unsupported command in 'additional-files': {entry}")

		#
		# elif parts[0] == "global-include":
		# 	for include_pat in parts[1:]:
		# 		for include_file in self.project_dir.rglob(include_pat):
		# 			if include_file.is_file():
		# 				copy_file(filename=include_file)
		#
		# elif parts[0] == "global-exclude":
		# 	for exclude_pat in parts[1:]:
		# 		for exclude_file in self.project_dir.rglob(exclude_pat):
		# 			if exclude_file.is_file():
		# 				exclude_file.unlink()
		# 				self.report_removed(exclude_file)

		#
		# elif parts[0] == "graft":
		# 	for graft_dir in self.project_dir.rglob(parts[1]):
		# 		for graft_file in graft_dir.rglob("*.*"):
		# 			if graft_file.is_file():
		# 				copy_file(graft_file)
		#
		# elif parts[0] == "prune":
		# 	for prune_dir in self.project_dir.rglob(parts[1]):
		# 		for prune_file in prune_dir.rglob("*.*"):
		# 			if prune_file.is_file():
		# 				prune_file.unlink()
		# 				self.report_removed(exclude_file)

		# pylint: enable=loop-invariant-statement
		return

	def write_license(self, dest_dir: PathPlus, dest_filename: str = "LICENSE") -> None:
		"""
		Write the ``LICENSE`` file.

		:param dest_dir: The directory to write the file into.
		:param dest_filename: The name of the file to write in ``dest_dir``.
		"""

		if self.config.get("license", None) is not None:
			target = dest_dir / dest_filename
			target.parent.maybe_make(parents=True)
			target.write_clean(self.config["license"].text)
			self.report_written(target)

	def parse_authors(self) -> Dict[str, str]:
		"""
		Parse the :tconf:`project.authors` and :tconf:`~project.maintainers` fields into :core-meta:`Author`,
		:core-meta:`Maintainer-Email` etc.

		:return: A mapping of field names to values.

			Possible field names are ``Author``, ``Author-Email``, ``Maintainer``, and ``Maintainer-Email``.
		"""  # noqa: D400

		author = []
		author_email = []
		maintainer = []
		maintainer_email = []

		for entry in self.config["authors"]:
			if entry["name"] and entry["email"]:
				address = Address(entry["name"], addr_spec=entry["email"])
				author_email.append(str(address))
			elif entry["email"]:
				author_email.append(entry["email"])
			elif entry["name"]:
				author.append(entry["name"])

		for entry in self.config["maintainers"]:
			if entry["name"] and entry["email"]:
				address = Address(entry["name"], addr_spec=entry["email"])
				author_email.append(str(address))
			elif entry["email"]:
				maintainer_email.append(entry["email"])
			elif entry["name"]:
				maintainer.append(entry["name"])

		# TODO: I'm not quite sure how PEP 621 expects a name for one author and the email for another to be handled.

		output = {}

		if author_email:
			output["Author-email"] = ", ".join(author_email)
		elif author:
			output["Author"] = word_join(author)

		if maintainer_email:
			output["Maintainer-email"] = ", ".join(maintainer_email)
		elif maintainer:
			output["Maintainer"] = word_join(maintainer)

		return output

	def get_metadata_map(self) -> MetadataMapping:
		"""
		Generate the content of the ``METADATA`` / ``PKG-INFO`` file.
		"""

		metadata_mapping = MetadataMapping()

		metadata_mapping["Metadata-Version"] = "2.2"
		metadata_mapping["Name"] = self.config["name"]
		metadata_mapping["Version"] = str(self.config["version"])

		def add_not_none(key: str, field: str) -> None:
			if self.config[key] is not None:
				metadata_mapping[field] = self.config[key]

		def add_multiple(key: str, field: str) -> None:
			for value in self.config[key]:
				metadata_mapping[field] = str(value)  # pylint: disable=loop-invariant-statement

		add_multiple("dynamic", "Dynamic")
		metadata_mapping.update(self.parse_authors())

		add_not_none("description", "Summary")
		add_not_none("license-key", "License")

		add_multiple("classifiers", "Classifier")
		add_multiple("dependencies", "Requires-Dist")

		if self.config["keywords"]:
			metadata_mapping["Keywords"] = ','.join(self.config["keywords"])

		seen_hp = False

		# pylint: disable=loop-invariant-statement
		for category, url in self.config["urls"].items():
			if category.lower() in {"homepage", "home page"} and not seen_hp:
				metadata_mapping["Home-page"] = url
				seen_hp = True
			else:
				metadata_mapping["Project-URL"] = f"{category}, {url}"

		for platform in (self.config.get("platforms", None) or ()):
			metadata_mapping["Platform"] = platform

		if self.config["requires-python"]:
			metadata_mapping["Requires-Python"] = str(self.config["requires-python"])

		for extra, requirements in self.config["optional-dependencies"].items():
			metadata_mapping["Provides-Extra"] = extra
			for requirement in requirements:
				requirement = ComparableRequirement(str(requirement))

				if requirement.marker:
					requirement.marker = f"({requirement.marker!s}) and extra == {extra!r}"
				else:
					requirement.marker = f"extra == {extra!r}"

				metadata_mapping["Requires-Dist"] = str(requirement)

		# pylint: enable=loop-invariant-statement

		# TODO:
		#  https://packaging.python.org/specifications/core-metadata/#requires-external-multiple-use
		#  https://packaging.python.org/specifications/core-metadata/#provides-dist-multiple-use
		#  https://packaging.python.org/specifications/core-metadata/#obsoletes-dist-multiple-use

		if self.config["readme"] is not None:
			metadata_mapping["Description"] = self.config["readme"].text
			metadata_mapping["Description-Content-Type"] = self.config["readme"].content_type

		return metadata_mapping

	def write_metadata(self, metadata_file: PathPlus, metadata_mapping: MetadataMapping) -> None:
		"""
		Write `Core Metadata`_ to the given file.

		.. _Core Metadata: https://packaging.python.org/specifications/core-metadata

		:param metadata_file:
		"""

		metadata_file.write_clean(metadata.dumps(metadata_mapping))
		self.report_written(metadata_file)

	def call_additional_hooks(self) -> None:
		"""
		Subclasses may call this method to give *their* subclasses an opportunity to run custom code.

		For example, the wheel builder calls this as the final step before adding files to the archive,
		giving an opportunity for subclasses of :class:`~.WheelBuilder` to include additional steps
		without having to override the entire :meth:`~.WheelBuilder.build_wheel` method.
		"""

	@abstractmethod
	def build(self) -> str:
		"""
		Build the distribution.

		:returns: The filename of the created archive.
		"""

		raise NotImplementedError


class SDistBuilder(AbstractBuilder):
	"""
	Builds source distributions using metadata read from ``pyproject.toml``.

	:param project_dir: The project to build the distribution for.
	:param build_dir: The (temporary) build directory.
	:default build_dir: :file:`{<project_dir>}/build/sdist`
	:param out_dir: The output directory.
	:default out_dir: :file:`{<project_dir>}/dist`
	:param verbose: Enable verbose output.
	"""

	@property
	def default_build_dir(self) -> PathPlus:  # pragma: no cover
		"""
		Provides a default for the ``build_dir`` argument.
		"""

		return self.project_dir / "build" / "sdist"

	@property
	def code_directory(self) -> str:
		"""
		The directory containing the code in the build and project directories.
		"""

		return ''

	def create_sdist_archive(self) -> str:
		"""
		Create the sdist archive.

		:return: The filename of the created archive.
		"""

		self.out_dir.maybe_make(parents=True)

		sdist_filename = self.out_dir / f"{self.archive_name}.tar.gz"
		with tarfile.open(sdist_filename, mode="w:gz", format=tarfile.PAX_FORMAT) as sdist_archive:
			for file in self.build_dir.rglob('*'):
				if file.is_file():
					arcname = posixpath_join(self.archive_name, file.relative_to(self.build_dir).as_posix())
					sdist_archive.add(str(file), arcname=arcname)

		self._echo(Fore.GREEN(f"Source distribution created at {sdist_filename.resolve().as_posix()}"))
		return os.path.basename(sdist_filename)

	def write_pyproject_toml(self) -> None:
		"""
		Write the ``pyproject.toml`` file.
		"""

		# Copy pyproject.toml
		pp_toml = dom_toml.load(self.project_dir / "pyproject.toml")

		# Ensure whey is the build backend and a requirement
		pp_toml.setdefault("build-system", {})
		current_requires = map(ComparableRequirement, pp_toml["build-system"].get("requires", ()))
		new_requirements = combine_requirements(ComparableRequirement("whey"), *current_requires)
		pp_toml["build-system"]["requires"] = list(map(str, sorted(new_requirements)))
		pp_toml["build-system"]["build-backend"] = "whey"

		dynamic = set(pp_toml["project"].get("dynamic", ()))

		# Make the "dependencies" static
		if "dependencies" in dynamic:
			dynamic.remove("dependencies")

			pp_toml["project"]["dependencies"] = list(map(str, sorted(self.config["dependencies"])))

		# Make the "classifiers" static
		if "classifiers" in dynamic:
			dynamic.remove("classifiers")

			pp_toml["project"]["classifiers"] = list(map(str, sorted(self.config["classifiers"])))

		# Make "requires-python" static
		if "requires-python" in dynamic:
			dynamic.remove("requires-python")

			pp_toml["project"]["requires-python"] = str(self.config["requires-python"])

		# Set the new value for "dynamic"
		pp_toml["project"]["dynamic"] = sorted(dynamic)

		dom_toml.dump(pp_toml, self.build_dir / "pyproject.toml", encoder=PyProjectTomlEncoder)
		self.report_copied(self.project_dir / "pyproject.toml", self.build_dir / "pyproject.toml")

	def build_sdist(self) -> str:
		"""
		Build the source distribution.

		:return: The filename of the created archive.
		"""

		if self.build_dir.is_dir():
			shutil.rmtree(self.build_dir)
		self.build_dir.maybe_make(parents=True)

		self.copy_source()
		self.copy_additional_files()
		self.write_license(self.build_dir)

		self.write_pyproject_toml()

		for filename in ["requirements.txt"]:
			source = self.project_dir / filename
			if source.is_file():
				dest = self.build_dir / filename
				dest.write_clean(source.read_text())
				self.report_copied(source, dest)

		self.write_readme()
		self.write_metadata(self.build_dir / "PKG-INFO", self.get_metadata_map())
		self.call_additional_hooks()

		return self.create_sdist_archive()

	def write_readme(self) -> None:
		"""
		Write the ``README.*`` file.
		"""

		if self.config["readme"] is None:
			return

		if self.config["readme"].content_type == "text/x-rst":
			target = self.build_dir / "README.rst"
		elif self.config["readme"].content_type == "text/markdown":
			target = self.build_dir / "README.md"
		else:
			target = self.build_dir / "README"

		target.parent.maybe_make(parents=True)
		target.write_clean(self.config["readme"].text)
		self.report_written(target)

	build = build_sdist


class WheelBuilder(AbstractBuilder):
	"""
	Builds wheel binary distributions using metadata read from ``pyproject.toml``.

	:param project_dir: The project to build the distribution for.
	:param build_dir: The (temporary) build directory.
	:default build_dir: :file:`{<project_dir>}/build/wheel`
	:param out_dir: The output directory.
	:default out_dir: :file:`{<project_dir>}/dist`
	:param verbose: Enable verbose output.

	.. autosummary-widths:: 11/32

	.. latex:vspace:: -10px
	"""

	@property
	def default_build_dir(self) -> PathPlus:  # pragma: no cover
		"""
		Provides a default for the ``build_dir`` argument.
		"""

		return self.project_dir / "build" / "wheel"

	@property
	def dist_info(self) -> PathPlus:
		"""
		The ``*.dist-info`` directory in the build directory.
		"""

		dist_info = self.build_dir / f"{self.archive_name}.dist-info"
		dist_info.maybe_make(parents=True)
		return dist_info

	@property
	def tag(self) -> str:
		"""
		The tag for the wheel.
		"""

		return "py3-none-any"

	@property
	def generator(self) -> str:
		"""
		The value for the ``Generator`` field in ``*.dist-info/WHEEL``.
		"""

		# this package
		from whey import __version__

		return f"whey ({__version__})"

	def write_entry_points(self) -> None:
		"""
		Write the list of entry points to the wheel, as specified in
		``[project.scripts]``, ``[project.gui-scripts]`` and ``[project.entry-points]``
		"""  # noqa: D400

		ep_dict = {}

		if self.config["scripts"]:
			ep_dict["console_scripts"] = self.config["scripts"]

		if self.config["gui-scripts"]:
			ep_dict["gui_scripts"] = self.config["gui-scripts"]

		ep_dict.update(self.config["entry-points"])

		entry_points_file = self.dist_info / "entry_points.txt"
		entry_points.dump(ep_dict, entry_points_file)
		self.report_written(entry_points_file)

	def write_wheel(self) -> None:
		"""
		Write the metadata to the ``WHEEL`` file.
		"""

		wheel_file = self.dist_info / "WHEEL"

		wheel_file.write_clean(
				wheel.dumps({
						"Wheel-Version": "1.0",
						"Generator": self.generator,
						"Root-Is-Purelib": True,
						"Tag": [self.tag],
						})
				)

		self.report_written(wheel_file)

	@staticmethod
	def get_source_epoch() -> Optional[datetime.datetime]:
		"""
		Returns the parsed value of the :envvar:`SOURCE_DATE_EPOCH` environment variable, or :py:obj:`None` if unset.

		See https://reproducible-builds.org/specs/source-date-epoch/ for the specification.

		:raises ValueError: if the value is in an invalid format.
		"""

		# If SOURCE_DATE_EPOCH is set (e.g. by Debian), it's used for timestamps inside the wheel.
		epoch: Optional[str] = os.environ.get("SOURCE_DATE_EPOCH")
		if epoch is None:
			return None
		elif epoch.isdigit() and sys.version_info >= (3, 11):
			return datetime.datetime.fromtimestamp(int(epoch), datetime.UTC)  # type: ignore[attr-defined]
		elif epoch.isdigit():
			return datetime.datetime.utcfromtimestamp(int(epoch))
		else:
			raise ValueError(f"'SOURCE_DATE_EPOCH' must be an integer with no fractional component, not {epoch!r}")

	def create_wheel_archive(self) -> str:
		"""
		Create the wheel archive.

		:return: The filename of the created archive.
		"""

		wheel_filename = self.out_dir / f"{self.archive_name}-{self.tag}.whl"
		self.out_dir.maybe_make(parents=True)

		mtime = self.get_source_epoch()

		non_record_filenames = []
		record_filenames = []

		for file in self.build_dir.rglob('*'):
			if not file.is_file():
				continue
			if "RECORD" in file.name and self.dist_info.name in file.parts:
				record_filenames.append(file)
				continue

			non_record_filenames.append(file)

		record_filenames = sort_paths(*record_filenames, self.dist_info / "RECORD")

		# Perhaps LZMA support in the future
		with handy_archives.ZipFile(wheel_filename, mode='w', compression=zipfile.ZIP_DEFLATED) as wheel_archive:
			with (self.dist_info / "RECORD").open('w') as fp:
				for file in sort_paths(*non_record_filenames):  # pylint: disable=loop-invariant-statement

					fp.write(f"{get_record_entry(file, relative_to=self.build_dir)}\n")

					wheel_archive.write_file(
							file,
							arcname=file.relative_to(self.build_dir),
							mtime=mtime,
							)

				for file in record_filenames:
					fp.write(f"{file.relative_to(self.build_dir).as_posix()},,\n")

			for file in record_filenames:
				wheel_archive.write_file(
						file,
						arcname=file.relative_to(self.build_dir),
						mtime=mtime,
						)
				self.report_written(file)

		self._echo(Fore.GREEN(f"Wheel created at {wheel_filename.resolve().as_posix()}"))

		return wheel_filename.name

	def create_editables_files(self) -> Iterator[ComparableRequirement]:
		"""
		Generate files with `editables`_ for use in a :pep:`660` wheel.

		.. _editables: https://pypi.org/project/editables/

		.. extras-require:: editable
			:scope: method
			:pyproject:

		:returns: An iterator of additional runtime requirements which should be added to the wheel's ``METADATA`` file.
		"""

		# this package
		from whey._editable import EditableProject

		pkgdir = self.project_dir / self.config["source-dir"] / self.config["package"]

		if not pkgdir.is_dir():
			message = f"Package directory {self.config['package']!r} not found"

			if self.config["source-dir"]:
				raise FileNotFoundError(f"{message} in {self.config['source-dir']!r}.")
			else:
				raise FileNotFoundError(f"{message}.")

		my_project = EditableProject(
				self.config["name"].replace('-', '_'),
				pkgdir.parent,
				)

		my_project.map_or_add_to_path(self.config["package"], pkgdir)

		for name, content in my_project.files():
			target = self.build_dir / name
			target.parent.maybe_make(parents=True)
			target.write_clean(content)
			self.report_written(target)

		yield from map(ComparableRequirement, my_project.dependencies())

	def build_wheel(self) -> str:
		"""
		Build the binary wheel distribution.

		:return: The filename of the created archive.
		"""

		if self.build_dir.is_dir():
			shutil.rmtree(self.build_dir)

		self.build_dir.maybe_make(parents=True)

		self.copy_source()
		self.copy_additional_files()
		self.write_license(self.dist_info)
		self.write_entry_points()
		self.write_metadata(self.dist_info / "METADATA", self.get_metadata_map())
		self.write_wheel()
		self.call_additional_hooks()

		return self.create_wheel_archive()

	build = build_wheel

	def build_editable(self) -> str:
		"""
		Build an editable wheel.

		An "editable" wheel uses the wheel format not for distribution but as ephemeral communication
		between the build system and the front end.
		This avoids having the build backend install anything directly.
		This wheel must not be exposed to end users, nor cached, nor distributed.

		You should use a different ``build_dir`` and ``out_dir`` to those used for standard wheel builds.

		The default implementation of this method does not call
		:meth:`~.AbstractBuilder.copy_source` or :meth:`~.AbstractBuilder.copy_additional_files`.

		.. extras-require:: editable
			:scope: method
			:pyproject:

		:return: The filename of the created archive.
		"""

		if self.build_dir.is_dir():
			shutil.rmtree(self.build_dir)

		self.build_dir.maybe_make(parents=True)

		extra_deps = self.create_editables_files()
		self.write_license(self.dist_info)
		self.write_entry_points()

		metadata = self.get_metadata_map()

		for dep in extra_deps:
			if not any(dep.name in req for req in metadata.get_all("Requires-Dist", ())):
				# Additional runtime requirement for editable wheels.
				metadata["Requires-Dist"] = str(dep)

		# Prevents uploading to PyPI, which you shouldn't do with an editable wheel.
		metadata["Classifier"] = "Private :: Do Not Upload"

		self.write_metadata(self.dist_info / "METADATA", metadata)
		self.write_wheel()
		self.call_additional_hooks()

		return self.create_wheel_archive()



#!/usr/bin/env python3
#
#  pep621.py
"""
:pep:`621` configuration parser.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Dict, cast

# 3rd party
import pyproject_parser.parsers
from dom_toml.parser import TOML_TYPES, BadConfigError
from domdf_python_tools.words import Plural, word_join
from pyproject_parser.type_hints import ProjectDict

__all__ = ("PEP621Parser", )

_field = Plural("field", "fields")


class PEP621Parser(pyproject_parser.parsers.PEP621Parser, inherit_defaults=True):
	"""
	Parser for :pep:`621` metadata from ``pyproject.toml``.
	"""

	defaults = {
			"description": None,
			"readme": None,
			"requires-python": None,
			"license": None,
			}

	def _parse(
			self,
			config: Dict[str, TOML_TYPES],
			set_defaults: bool = False,
			) -> ProjectDict:

		dynamic_fields = config.get("dynamic", [])
		parsed_config = {"dynamic": dynamic_fields}

		for key in self.keys:

			if key in config and key in dynamic_fields:
				raise BadConfigError(f"{key!r} was listed in 'project.dynamic' but a value was given.")
			elif key not in config:
				# Ignore absent values
				pass
			elif hasattr(self, f"parse_{key.replace('-', '_')}"):
				parsed_config[key] = getattr(self, f"parse_{key.replace('-', '_')}")(config)
			elif key in config:
				parsed_config[key] = config[key]

		if set_defaults:
			for key, value in self.defaults.items():
				parsed_config.setdefault(key, value)

			for key, factory in self.factories.items():
				value = factory()
				parsed_config.setdefault(key, value)

		return cast(ProjectDict, parsed_config)

	def parse(  # type: ignore[override]
		self,
		config: Dict[str, TOML_TYPES],
		set_defaults: bool = False,
		) -> ProjectDict:
		"""
		Parse the TOML configuration.

		:param config:
		:param set_defaults: If :py:obj:`True`, the values in
			:attr:`dom_toml.parser.AbstractConfigParser.defaults` and
			:attr:`dom_toml.parser.AbstractConfigParser.factories`
			will be set as defaults for the returned mapping.
		"""

		dynamic_fields = set(config.get("dynamic", []))

		if "name" in dynamic_fields:
			raise BadConfigError("The 'project.name' field may not be dynamic.")
		elif "name" not in config:
			raise BadConfigError("The 'project.name' field must be provided.")

		if dynamic_fields:
			# TODO: Support the remaining fields as dynamic
			# TODO: dynamic version numbers by parsing AST for __version__ in __init__.py

			supported_dynamic = {"classifiers", "requires-python", "dependencies"}
			unsupported_fields = dynamic_fields - supported_dynamic

			if unsupported_fields:
				supported = word_join(sorted(supported_dynamic), oxford=True, use_repr=True)
				unsupported = word_join(sorted(unsupported_fields), oxford=True, use_repr=True)
				raise BadConfigError(
						f"Unsupported dynamic {_field(len(unsupported_fields))} {unsupported}.\n"
						f"note: whey only supports {supported} as dynamic fields."
						)

		if "version" not in config:
			raise BadConfigError("The 'project.version' field must be provided.")

		return self._parse(config, set_defaults)



#!/usr/bin/env python
#
#  additional_files.py
"""
Parser for the ``additional-files`` option.
"""
#
#  Copyright © 2020-2023 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import abc
from typing import Any, Dict, Iterable, Iterator, List, Optional
from warnings import warn

# 3rd party
import attr
from dom_toml.parser import BadConfigError
from domdf_python_tools.paths import PathPlus, sort_paths

__all__ = ["AdditionalFilesEntry", "Exclude", "Include", "RecursiveExclude", "RecursiveInclude", "from_entry"]


class AdditionalFilesEntry(abc.ABC):
	"""
	An abstract command in ``additional-files``.
	"""

	@classmethod
	@abc.abstractmethod
	def parse(cls, parameters: str) -> "AdditionalFilesEntry":
		"""
		Parse the command's parameters.

		:param parameters:
		"""

		raise NotImplementedError

	@abc.abstractmethod
	def iter_files(self, directory: PathPlus) -> Iterator[PathPlus]:
		"""
		Returns an iterator over files to be included or excluded by this command.

		:param directory: The project or build directory.
		"""

		raise NotImplementedError

	@abc.abstractmethod
	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of the command entry.
		"""

		raise NotImplementedError


def _to_list(_: Iterable[str]) -> List[str]:
	return list(_)


@attr.define
class Include(AdditionalFilesEntry):
	"""
	Include a single file, or multiple files with a pattern.
	"""

	#: Glob patterns (with complete paths from the project root)
	patterns: List[str] = attr.field(converter=_to_list)

	@classmethod
	def parse(cls, parameters: str) -> "Include":
		"""
		Parse the command's parameters.

		:param parameters:
		"""

		if not parameters:
			raise BadConfigError(f"additional-files: 'include' must have at least one path or pattern specified.")

		return cls(parameters.split(' '))

	def iter_files(self, directory: PathPlus) -> Iterator[PathPlus]:
		"""
		Returns an iterator over files to be included by this command.

		:param directory: The project directory.
		"""

		for include_pat in self.patterns:
			for include_file in sorted(directory.glob(include_pat)):
				if include_file.is_file():
					yield include_file

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of the command entry.
		"""

		return {
				"command": "include",
				**attr.asdict(self),
				}


@attr.define
class Exclude(AdditionalFilesEntry):
	"""
	Exclude a single file, or multiple files with a pattern.
	"""

	#: Glob patterns (with complete paths from the project root)
	patterns: List[str] = attr.field(converter=_to_list)

	@classmethod
	def parse(cls, parameters: str) -> "Exclude":
		"""
		Parse the command's parameters.

		:param parameters:
		"""

		if not parameters:
			raise BadConfigError(f"additional-files: 'exclude' must have at least one path or pattern specified.")

		return cls(parameters.split(' '))

	def iter_files(self, directory: PathPlus) -> Iterator[PathPlus]:
		"""
		Returns an iterator over files to be excluded by this command.

		:param directory: The build directory.
		"""

		for exclude_pat in self.patterns:
			for exclude_file in sorted(directory.glob(exclude_pat)):
				if exclude_file.is_file():
					yield exclude_file

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of the command entry.
		"""

		return {
				"command": "exclude",
				**attr.asdict(self),
				}


@attr.define
class RecursiveInclude(AdditionalFilesEntry):
	"""
	Recursively include files in a directory based on patterns.
	"""

	#: The directory to start from.
	path: str

	#: Glob patterns.
	patterns: List[str] = attr.field(converter=_to_list)

	@classmethod
	def parse(cls, parameters: str) -> "RecursiveInclude":
		"""
		Parse the command's parameters.

		:param parameters:
		"""

		parts = parameters.split(' ')
		if len(parts) < 2:
			raise BadConfigError(
					f"additional-files: 'recursive-include' must have one path and at least one pattern specified."
					)

		return cls(parts[0], parts[1:])

	def iter_files(self, directory: PathPlus) -> Iterator[PathPlus]:
		"""
		Returns an iterator over files to be included by this command.

		:param directory: The project directory.
		"""

		for include_pat in self.patterns:
			for include_file in sort_paths(*(directory / self.path).rglob(include_pat)):
				if "__pycache__" in include_file.parts:
					continue

				if include_file.is_file():
					yield include_file

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of the command entry.
		"""

		return {
				"command": "recursive-include",
				**attr.asdict(self),
				}


@attr.define
class RecursiveExclude(AdditionalFilesEntry):
	"""
	Recursively exclude files in a directory based on patterns.
	"""

	#: The directory to start from.
	path: str

	#: Glob patterns.
	patterns: List[str] = attr.field(converter=_to_list)

	@classmethod
	def parse(cls, parameters: str) -> "RecursiveExclude":
		"""
		Parse the command's parameters.

		:param parameters:
		"""

		parts = parameters.split(' ')
		if len(parts) < 2:
			raise BadConfigError(
					f"additional-files: 'recursive-exclude' must have one path and at least one pattern specified."
					)

		return cls(parts[0], parts[1:])

	def iter_files(self, directory: PathPlus) -> Iterator[PathPlus]:
		"""
		Returns an iterator over files to be excluded by this command.

		:param directory: The build directory.
		"""
		for exclude_pat in self.patterns:
			for exclude_file in sort_paths(*(directory / self.path).rglob(exclude_pat)):
				if exclude_file.is_file():
					yield exclude_file

	def to_dict(self) -> Dict[str, Any]:
		"""
		Returns a dictionary representation of the command entry.
		"""

		return {
				"command": "recursive-exclude",
				**attr.asdict(self),
				}


def from_entry(line: str) -> Optional[AdditionalFilesEntry]:
	"""
	Parse a `MANIFEST.in`_-style entry.

	.. _MANIFEST.in: https://packaging.python.org/guides/using-manifest-in/

	:param line:

	:returns: An :class:`~.AdditionalFilesEntry` for known commands,
		or :py:obj:`None` if an unknown command is found in the entry.
	"""

	command, *parameters = line.split(' ')
	parameter_string = ' '.join(parameters)

	if command == "include":
		return Include.parse(parameter_string)
	elif command == "exclude":
		return Exclude.parse(parameter_string)
	elif command == "recursive-include":
		return RecursiveInclude.parse(parameter_string)
	elif command == "recursive-exclude":
		return RecursiveExclude.parse(parameter_string)
	else:  # pragma: no cover
		warn(f"Unsupported command in 'additional-files': {line}")
		return None



#!/usr/bin/env python
#
#  _editable.py
"""
Internal plumbing for editable installs.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  `map` method adapted from https://github.com/pfmoore/editables
#  Copyright © 2020 Paul Moore
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from pathlib import Path
from typing import Dict, List

# 3rd party
import editables  # nodep
from domdf_python_tools.typing import PathLike

__all__ = ("EditableProject", )


class EditableProject(editables.EditableProject):  # noqa: D101
	project_name: str
	project_dir: Path
	redirections: Dict[str, str]
	path_entries: List[Path]

	def map(self, name: str, target: PathLike) -> None:  # noqa: A003,D102  # pylint: disable=redefined-builtin
		if '.' in name:
			raise editables.EditableException(f"Cannot map {name} as it is not a top-level package")

		abs_target = self.make_absolute(target)

		if abs_target.is_dir():
			abs_target = abs_target / "__init__.py"

		if abs_target.is_file():
			self.redirections[name] = abs_target.as_posix()
		else:
			raise editables.EditableException(f"{target} is not a valid Python package or module")

	def map_or_add_to_path(self, name: str, pkgdir: Path) -> None:
		"""
		``.map`` a conventional package, or ``.add_to_path`` a :pep:`420` namespace package.

		:param name: The package name.
		:param pkgdir: The package directory.
		"""

		if (pkgdir / "__init__.py").is_file():
			self.map(name, pkgdir)
		else:
			# namespace package
			self.add_to_path(pkgdir.parent)



#!/usr/bin/env python3
#
#  __init__.py
"""
A simple Python wheel builder for simple projects.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import os

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2021 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "0.1.1"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = ("build_sdist", "build_wheel")


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None) -> str:  # noqa: MAN001
	"""
	:pep:`517` hook to build a wheel binary distribution.

	.. seealso:: https://www.python.org/dev/peps/pep-0517/#build-wheel

	:param wheel_directory:
	:param config_settings:
	:param metadata_directory:

	:returns: The filename of the created archive.
	"""

	# 3rd party
	from consolekit.tracebacks import handle_tracebacks
	from domdf_python_tools.paths import PathPlus, TemporaryPathPlus

	# this package
	from whey.foreman import Foreman
	from whey.utils import WheyBackendTBHandler

	show_traceback = bool(int(os.getenv("WHEY_TRACEBACK", 0)))
	verbose = bool(int(os.getenv("WHEY_VERBOSE", 1)))

	with TemporaryPathPlus() as tmpdir, handle_tracebacks(show_traceback, WheyBackendTBHandler):
		foreman = Foreman(project_dir=PathPlus.cwd())
		return foreman.build_wheel(build_dir=tmpdir, out_dir=wheel_directory, verbose=verbose)


def build_sdist(sdist_directory, config_settings=None) -> str:  # noqa: MAN001
	"""
	:pep:`517` hook to build a source distribution.

	.. seealso:: https://www.python.org/dev/peps/pep-0517/#build-sdist

	:param sdist_directory:
	:param config_settings:

	:returns: The filename of the created archive.
	"""

	# 3rd party
	from consolekit.tracebacks import handle_tracebacks
	from domdf_python_tools.paths import PathPlus, TemporaryPathPlus

	# this package
	from whey.foreman import Foreman
	from whey.utils import WheyBackendTBHandler

	show_traceback = bool(int(os.getenv("WHEY_TRACEBACK", 0)))
	verbose = bool(int(os.getenv("WHEY_VERBOSE", 1)))

	with TemporaryPathPlus() as tmpdir, handle_tracebacks(show_traceback, WheyBackendTBHandler):
		foreman = Foreman(project_dir=PathPlus.cwd())
		return foreman.build_sdist(build_dir=tmpdir, out_dir=sdist_directory, verbose=verbose)


def get_requires_for_build_sdist(config_settings=None):  # pragma: no cover  # noqa: MAN001,MAN002
	return []


def get_requires_for_build_editable(config_settings=None):  # pragma: no cover  # noqa: MAN001,MAN002
	return ["editables>=0.2"]


def build_editable(wheel_directory, config_settings=None, metadata_directory=None):  # noqa: MAN001,MAN002
	"""
	:pep:`517` hook to build an editable wheel.

	.. seealso:: :pep:`660`

	:param wheel_directory:
	:param config_settings:
	:param metadata_directory:
	"""

	# stdlib
	from typing import Type, cast

	# 3rd party
	from consolekit.tracebacks import handle_tracebacks
	from domdf_python_tools.paths import PathPlus, TemporaryPathPlus

	# this package
	from whey.builder import WheelBuilder
	from whey.foreman import Foreman
	from whey.utils import WheyBackendTBHandler

	show_traceback = bool(int(os.getenv("WHEY_TRACEBACK", 0)))
	verbose = bool(int(os.getenv("WHEY_VERBOSE", 1)))

	with TemporaryPathPlus() as tmpdir, handle_tracebacks(show_traceback, WheyBackendTBHandler):
		foreman = Foreman(project_dir=PathPlus.cwd())
		builder_cls: Type[WheelBuilder] = cast(Type[WheelBuilder], foreman.get_builder("wheel"))
		builder = builder_cls(
				foreman.project_dir,
				foreman.config,
				build_dir=tmpdir,
				out_dir=wheel_directory,
				verbose=verbose,
				)
		return builder.build_editable()



#!/usr/bin/env python3
#
#  __main__.py
"""
A simple Python wheel builder for simple projects.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import sys

# 3rd party
import click
from consolekit import click_command
from consolekit.options import (
		DescribedArgument,
		auto_default_argument,
		auto_default_option,
		colour_option,
		flag_option,
		version_option
		)
from consolekit.tracebacks import handle_tracebacks

if False:  # TYPE_CHECKING:  # pylint: disable=using-constant-test
	# stdlib
	from typing import Iterable, Optional

	# 3rd party
	from consolekit.terminal_colours import ColourTrilean
	from domdf_python_tools.typing import PathLike

__all__ = ("main", )


def version_callback(ctx: click.Context, param: click.Option, value: int) -> None:
	"""
	Callback for displaying the package version (and optionally the Python runtime).
	"""

	# 3rd party
	from consolekit.versions import get_version_callback

	# this package
	import whey

	key_dependencies = ["dist-meta", "handy-archives", "packaging", "pyproject-parser", "shippinglabel"]
	return get_version_callback(whey.__version__, "whey", key_dependencies)(ctx, param, value)


@version_option(version_callback)
@flag_option(
		"-T",
		"--traceback",
		"show_traceback",
		help="Show the complete traceback on error.",
		envvar="WHEY_TRACEBACK",
		)
@colour_option()
@flag_option("-S", "--show-builders", help="Show the builders which will be used, and exit.")
@flag_option("-v", "--verbose", help="Enable verbose output.", envvar="WHEY_VERBOSE")
@auto_default_option(
		"-o",
		"--out-dir",
		type=click.STRING,
		help="The output directory.",
		metavar="DIRECTORY",
		)
@auto_default_option(
		"--build-dir",
		type=click.STRING,
		help="The temporary build directory.",
		metavar="DIRECTORY",
		)
@auto_default_option(
		"-B",
		"--builder",
		type=click.STRING,
		help="The builder to build with.",
		multiple=True,
		metavar="BUILDER",
		)
@flag_option("-b", "--binary", help="Build a binary distribution.")
@flag_option("-w", "--wheel", help="Build a wheel.")
@flag_option("-s", "--sdist", help="Build a source distribution.")
@auto_default_argument(
		"project",
		type=click.STRING,
		cls=DescribedArgument,
		description="The path to the project to build.",
		)
@click_command()
def main(
		project: "PathLike" = '.',
		build_dir: "Optional[str]" = None,
		out_dir: "Optional[str]" = None,
		sdist: bool = False,
		wheel: bool = False,
		binary: bool = False,
		builder: "Optional[Iterable[str]]" = None,
		verbose: bool = False,
		colour: "ColourTrilean" = None,
		show_traceback: bool = False,
		show_builders: bool = False,
		) -> None:
	"""
	Build a wheel for the given project.
	"""

	# 3rd party
	from domdf_python_tools.paths import PathPlus

	# this package
	from whey.foreman import Foreman
	from whey.utils import WheyTracebackHandler, parse_custom_builders, print_builder_names

	if not any((binary, sdist, wheel, builder)):
		wheel = True
		sdist = True

	project = PathPlus(project).resolve()

	with handle_tracebacks(show_traceback, WheyTracebackHandler):
		foreman = Foreman(project_dir=project)

		custom_builders = parse_custom_builders(builder)

		if verbose or show_builders:
			print_builder_names(foreman, custom_builders, sdist=sdist, wheel=wheel, binary=binary)

		if show_builders:
			sys.exit(0)

		click.echo(f"Building {foreman.project_dir.as_posix()}")

		if wheel:
			foreman.build_wheel(
					build_dir=build_dir,
					out_dir=out_dir,
					verbose=verbose,
					colour=colour,
					)

		if sdist:
			foreman.build_sdist(
					build_dir=build_dir,
					out_dir=out_dir,
					verbose=verbose,
					colour=colour,
					)

		if binary:
			foreman.build_binary(
					build_dir=build_dir,
					out_dir=out_dir,
					verbose=verbose,
					colour=colour,
					)

		for custom_builder in custom_builders.values():
			custom_builder(
					foreman.project_dir,
					foreman.config,
					build_dir=build_dir,
					out_dir=out_dir,
					verbose=verbose,
					colour=colour,
					).build()


if __name__ == "__main__":
	sys.exit(main())
